use crate::{
    bundle::{BundleType, BundleValue, DynBundle},
    firrtl,
    intern::Interned,
    module::Module,
};
use clap::{Args, Parser, Subcommand, ValueHint};
use std::{fmt, path::PathBuf};

enum CliErrorImpl {
    IoError(std::io::Error),
}

pub type Result<T = (), E = CliError> = std::result::Result<T, E>;

pub struct CliError(CliErrorImpl);

impl fmt::Debug for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match &self.0 {
            CliErrorImpl::IoError(e) => e.fmt(f),
        }
    }
}

impl fmt::Display for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match &self.0 {
            CliErrorImpl::IoError(e) => e.fmt(f),
        }
    }
}

impl std::error::Error for CliError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match &self.0 {
            CliErrorImpl::IoError(e) => e.source(),
        }
    }
}

impl From<std::io::Error> for CliError {
    fn from(value: std::io::Error) -> Self {
        CliError(CliErrorImpl::IoError(value))
    }
}

pub trait RunPhase<Arg> {
    fn run(self, arg: Arg) -> Result;
}

#[derive(Args)]
#[non_exhaustive]
pub struct FirrtlArgs {
    /// the directory to put the generated FIRRTL and associated files in
    #[arg(short, long, value_hint = ValueHint::DirPath)]
    pub output: PathBuf,
    /// the stem of the generated .fir file, e.g. to get foo.fir, pass --file-stem=foo
    #[arg(long)]
    pub file_stem: Option<String>,
}

impl FirrtlArgs {
    fn run_impl(self, top_module: Module<DynBundle>) -> Result {
        firrtl::export(
            firrtl::FileBackend {
                dir_path: self.output,
                top_fir_file_stem: self.file_stem,
            },
            &top_module,
        )?;
        Ok(())
    }
}

impl<T: BundleValue> RunPhase<Module<T>> for FirrtlArgs
where
    T::Type: BundleType<Value = T>,
{
    fn run(self, top_module: Module<T>) -> Result {
        self.run_impl(top_module.canonical())
    }
}

impl<T: BundleValue> RunPhase<Interned<Module<T>>> for FirrtlArgs
where
    T::Type: BundleType<Value = T>,
{
    fn run(self, top_module: Interned<Module<T>>) -> Result {
        self.run(*top_module)
    }
}

#[derive(Subcommand)]
enum CliCommand {
    /// Generate FIRRTL
    Firrtl(FirrtlArgs),
}

/// a simple CLI
///
/// Use like:
///
/// ```no_run
/// # use fayalite::hdl_module;
/// # #[hdl_module]
/// # fn my_module() {}
/// use fayalite::cli;
///
/// fn main() -> cli::Result {
///     cli::Cli::parse().run(my_module())
/// }
/// ```
///
/// You can also use it with a larger [`clap`]-based CLI like so:
///
/// ```no_run
/// # use fayalite::hdl_module;
/// # #[hdl_module]
/// # fn my_module() {}
/// use clap::{Subcommand, Parser};
/// use fayalite::cli;
///
/// #[derive(Subcommand)]
/// pub enum Cmd {
///     #[command(flatten)]
///     Fayalite(cli::Cli),
///     MySpecialCommand {
///         #[arg(long)]
///         foo: bool,
///     },
/// }
///
/// #[derive(Parser)]
/// pub struct Cli {
///     #[command(subcommand)]
///     cmd: Cmd, // or just use cli::Cli directly if you don't need more subcommands
/// }
///
/// fn main() -> cli::Result {
///     match Cli::parse().cmd {
///         Cmd::Fayalite(v) => v.run(my_module())?,
///         Cmd::MySpecialCommand { foo } => println!("special: foo={foo}"),
///     }
///     Ok(())
/// }
/// ```
#[derive(Parser)]
// clear things that would be crate-specific
#[command(name = "Fayalite Simple CLI", about = None, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    subcommand: CliCommand,
}

impl clap::Subcommand for Cli {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands(cmd)
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands_for_update(cmd)
    }

    fn has_subcommand(name: &str) -> bool {
        CliCommand::has_subcommand(name)
    }
}

impl<T> RunPhase<T> for Cli
where
    FirrtlArgs: RunPhase<T>,
{
    fn run(self, arg: T) -> Result {
        match self.subcommand {
            CliCommand::Firrtl(firrtl_args) => firrtl_args.run(arg),
        }
    }
}

impl Cli {
    /// forwards to [`clap::Parser::parse()`] so you don't have to import [`clap::Parser`]
    pub fn parse() -> Self {
        clap::Parser::parse()
    }
    /// forwards to [`RunPhase::run()`] so you don't have to import [`RunPhase`]
    pub fn run<T>(self, top_module: T) -> Result
    where
        Self: RunPhase<T>,
    {
        RunPhase::run(self, top_module)
    }
}
