// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
#![allow(clippy::type_complexity)]
#![allow(clippy::multiple_bound_locations)]
use crate::{
    annotations::{Annotation, IntoAnnotations, TargetedAnnotation},
    array::{Array, ArrayType, ArrayTypeTrait, ValueArrayOrSlice},
    bundle::{BundleType, BundleValue, DynBundle, DynBundleType},
    clock::{Clock, ClockType},
    expr::{Expr, ExprEnum, ExprTrait, Flow, ToExpr},
    int::{DynUInt, DynUIntType, UInt, UIntType},
    intern::{Intern, Interned},
    module::ScopedNameId,
    source_location::SourceLocation,
    ty::{AsMask, DynCanonicalType, DynCanonicalValue, DynType, Type, Value},
    util::DebugAsDisplay,
};
use bitvec::slice::BitSlice;
use std::{
    cell::RefCell,
    fmt,
    hash::{Hash, Hasher},
    num::NonZeroU32,
    rc::Rc,
};

#[derive(Value, Clone, PartialEq, Eq, Hash, Debug)]
pub struct ReadStruct<Element> {
    pub addr: DynUInt,
    pub en: UInt<1>,
    pub clk: Clock,
    #[hdl(flip)]
    pub data: Element,
}

#[derive(Value, Clone, PartialEq, Eq, Hash, Debug)]
pub struct WriteStruct<Element: Value> {
    pub addr: DynUInt,
    pub en: UInt<1>,
    pub clk: Clock,
    pub data: Element,
    pub mask: AsMask<Element>,
}

#[allow(clippy::multiple_bound_locations)]
#[derive(Value, Clone, PartialEq, Eq, Hash, Debug)]
pub struct ReadWriteStruct<Element: Value> {
    pub addr: DynUInt,
    pub en: UInt<1>,
    pub clk: Clock,
    #[hdl(flip)]
    pub rdata: Element,
    pub wmode: UInt<1>,
    pub wdata: Element,
    pub wmask: AsMask<Element>,
}

mod sealed {
    pub trait Sealed {}
}

pub trait PortType:
    sealed::Sealed + Clone + Eq + Hash + fmt::Debug + Send + Sync + 'static
{
    type PortKindTy: Copy + Eq + Hash + fmt::Debug + Send + Sync + 'static;
    type PortType: BundleType<Value = Self::PortValue>;
    type PortValue: BundleValue<Type = Self::PortType>;
    fn port_kind(port_kind: Self::PortKindTy) -> PortKind;
    fn from_port_kind(port_kind: PortKind) -> Self::PortKindTy;
    fn port_ty(port: &MemPort<Self>) -> Self::PortType;
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DynPortType {
    kind: PortKind,
}

impl sealed::Sealed for DynPortType {}

impl PortType for DynPortType {
    type PortKindTy = PortKind;
    type PortType = DynBundleType;
    type PortValue = DynBundle;

    fn port_kind(port_kind: Self::PortKindTy) -> PortKind {
        port_kind
    }

    fn from_port_kind(port_kind: PortKind) -> Self::PortKindTy {
        port_kind
    }

    fn port_ty(port: &MemPort<Self>) -> Self::PortType {
        match port.port_kind {
            PortKind::ReadOnly => MemPort::<ReadStruct<DynCanonicalValue>>::from_canonical(*port)
                .ty()
                .canonical(),
            PortKind::WriteOnly => MemPort::<WriteStruct<DynCanonicalValue>>::from_canonical(*port)
                .ty()
                .canonical(),
            PortKind::ReadWrite => {
                MemPort::<ReadWriteStruct<DynCanonicalValue>>::from_canonical(*port)
                    .ty()
                    .canonical()
            }
        }
    }
}

pub trait PortStruct:
    BundleValue
    + sealed::Sealed
    + PortType<PortKindTy = (), PortType = <Self as ToExpr>::Type, PortValue = Self>
where
    Self::Type: BundleType<Value = Self>,
{
    type Element: Value<Type = Self::ElementType>;
    type ElementType: Type<Value = Self::Element>;
    const PORT_KIND: PortKind;

    fn addr(this: Expr<Self>) -> Expr<DynUInt>;
    fn en(this: Expr<Self>) -> Expr<UInt<1>>;
    fn clk(this: Expr<Self>) -> Expr<Clock>;
    fn rdata(this: Expr<Self>) -> Option<Expr<Self::Element>>;
    fn wdata(this: Expr<Self>) -> Option<Expr<Self::Element>>;
    fn wmask(this: Expr<Self>) -> Option<Expr<AsMask<Self::Element>>>;
    fn wmode(this: Expr<Self>) -> Expr<UInt<1>>;
}

macro_rules! impl_port_struct {
    (
        impl<$Element:ident> _ for $Struct:ident {
            fn port_ty($port_ty_port:ident: $port_ty_port_ty:ty) -> $port_ty_ret_ty:ty {
                $($port_ty_body:tt)*
            }

            $($body:tt)*
        }
    ) => {
        impl<$Element: Value> sealed::Sealed for $Struct<$Element>
        where
            $Element::Type: Type<Value = $Element>,
        {
        }

        impl<$Element: Value> PortType for $Struct<$Element>
        where
            $Element::Type: Type<Value = $Element>,
        {
            type PortKindTy = ();
            type PortType = <Self as ToExpr>::Type;
            type PortValue = Self;

            fn port_kind(_port_kind: Self::PortKindTy) -> PortKind {
                Self::PORT_KIND
            }

            fn from_port_kind(port_kind: PortKind) -> Self::PortKindTy {
                assert_eq!(port_kind, Self::PORT_KIND);
            }

            fn port_ty($port_ty_port: $port_ty_port_ty) -> $port_ty_ret_ty {
                $($port_ty_body)*
            }
        }

        impl<$Element: Value> PortStruct for $Struct<$Element>
        where
            $Element::Type: Type<Value = $Element>,
        {
            type Element = $Element;
            type ElementType = $Element::Type;

            fn addr(this: Expr<Self>) -> Expr<DynUInt> {
                this.addr
            }

            fn en(this: Expr<Self>) -> Expr<UInt<1>> {
                this.en
            }

            fn clk(this: Expr<Self>) -> Expr<Clock> {
                this.clk
            }

            $($body)*
        }
    };
}

impl_port_struct! {
    impl<Element> _ for ReadStruct {
        fn port_ty(port: &MemPort<Self>) -> <MemPort<Self> as ToExpr>::Type {
            type T<V> = <V as ToExpr>::Type;
            T::<Self> {
                addr: port.addr_type,
                en: UIntType::new(),
                clk: ClockType,
                data: Element::Type::from_dyn_canonical_type(port.mem_element_type),
            }
        }

        const PORT_KIND: PortKind = PortKind::ReadOnly;

        fn rdata(this: Expr<Self>) -> Option<Expr<Self::Element>> {
            Some(this.data)
        }

        fn wdata(_this: Expr<Self>) -> Option<Expr<Self::Element>> {
            None
        }

        fn wmask(_this: Expr<Self>) -> Option<Expr<AsMask<Self::Element>>> {
            None
        }

        fn wmode(_this: Expr<Self>) -> Expr<UInt<1>> {
            false.to_expr()
        }
    }
}

impl_port_struct! {
    impl<Element> _ for WriteStruct {
        fn port_ty(port: &MemPort<Self>) -> <MemPort<Self> as ToExpr>::Type {
            type T<V> = <V as ToExpr>::Type;
            let element_ty = Element::Type::from_dyn_canonical_type(port.mem_element_type);
            T::<Self> {
                addr: port.addr_type,
                en: UIntType::new(),
                clk: ClockType,
                mask: element_ty.mask_type(),
                data: element_ty,
            }
        }

        const PORT_KIND: PortKind = PortKind::WriteOnly;

        fn rdata(_this: Expr<Self>) -> Option<Expr<Self::Element>> {
            None
        }

        fn wdata(this: Expr<Self>) -> Option<Expr<Self::Element>> {
            Some(this.data)
        }

        fn wmask(this: Expr<Self>) -> Option<Expr<AsMask<Self::Element>>> {
            Some(this.mask)
        }

        fn wmode(_this: Expr<Self>) -> Expr<UInt<1>> {
            true.to_expr()
        }
    }
}

impl_port_struct! {
    impl<Element> _ for ReadWriteStruct {
        fn port_ty(port: &MemPort<Self>) -> <MemPort<Self> as ToExpr>::Type {
            type T<V> = <V as ToExpr>::Type;
            let element_ty = Element::Type::from_dyn_canonical_type(port.mem_element_type);
            T::<Self> {
                addr: port.addr_type,
                en: UIntType::new(),
                clk: ClockType,
                rdata: element_ty.clone(),
                wmode: UIntType::new(),
                wmask: element_ty.mask_type(),
                wdata: element_ty,
            }
        }

        const PORT_KIND: PortKind = PortKind::ReadWrite;

        fn rdata(this: Expr<Self>) -> Option<Expr<Self::Element>> {
            Some(this.rdata)
        }

        fn wdata(this: Expr<Self>) -> Option<Expr<Self::Element>> {
            Some(this.wdata)
        }

        fn wmask(this: Expr<Self>) -> Option<Expr<AsMask<Self::Element>>> {
            Some(this.wmask)
        }

        fn wmode(this: Expr<Self>) -> Expr<UInt<1>> {
            this.wmode
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub enum PortKind {
    ReadOnly,
    WriteOnly,
    ReadWrite,
}

impl PortKind {
    pub const fn can_read(self) -> bool {
        match self {
            PortKind::ReadOnly => true,
            PortKind::WriteOnly => false,
            PortKind::ReadWrite => true,
        }
    }
    pub const fn can_write(self) -> bool {
        match self {
            PortKind::ReadOnly => false,
            PortKind::WriteOnly => true,
            PortKind::ReadWrite => true,
        }
    }
    pub const fn is_read_only(self) -> bool {
        matches!(self, PortKind::ReadOnly)
    }
    pub const fn is_write_only(self) -> bool {
        matches!(self, PortKind::WriteOnly)
    }
    pub const fn is_read_write(self) -> bool {
        matches!(self, PortKind::ReadWrite)
    }
    pub const fn port_name_prefix(self) -> &'static str {
        match self {
            PortKind::ReadOnly => "r",
            PortKind::WriteOnly => "w",
            PortKind::ReadWrite => "rw",
        }
    }
    pub const fn rdata_name(self) -> Option<&'static str> {
        match self {
            PortKind::ReadOnly => Some("data"),
            PortKind::WriteOnly => None,
            PortKind::ReadWrite => Some("rdata"),
        }
    }
    pub const fn wdata_name(self) -> Option<&'static str> {
        match self {
            PortKind::ReadOnly => None,
            PortKind::WriteOnly => Some("data"),
            PortKind::ReadWrite => Some("wdata"),
        }
    }
    pub const fn wmask_name(self) -> Option<&'static str> {
        match self {
            PortKind::ReadOnly => None,
            PortKind::WriteOnly => Some("mask"),
            PortKind::ReadWrite => Some("wmask"),
        }
    }
    pub const fn wmode_name(self) -> Option<&'static str> {
        match self {
            PortKind::ReadOnly => None,
            PortKind::WriteOnly => None,
            PortKind::ReadWrite => Some("wmode"),
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct PortName {
    pub kind: PortKind,
    pub index: usize,
}

impl fmt::Display for PortName {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}{}", self.kind.port_name_prefix(), self.index)
    }
}

impl fmt::Debug for PortName {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt::Display::fmt(self, f)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct MemPort<T: PortType> {
    mem_name: ScopedNameId,
    source_location: SourceLocation,
    port_kind: T::PortKindTy,
    port_index: usize,
    addr_type: DynUIntType,
    mem_element_type: Interned<dyn DynCanonicalType>,
}

impl<T: PortType> fmt::Debug for MemPort<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            mem_name,
            source_location: _,
            port_kind: _,
            port_index: _,
            addr_type,
            mem_element_type,
        } = self;
        f.debug_struct("MemPort")
            .field("mem_name", mem_name)
            .field("port_name", &self.port_name())
            .field("addr_type", addr_type)
            .field("mem_element_type", mem_element_type)
            .finish_non_exhaustive()
    }
}

impl<T: PortType> ToExpr for MemPort<T> {
    type Type = T::PortType;

    fn ty(&self) -> Self::Type {
        T::port_ty(self)
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: PortType> MemPort<T> {
    pub fn source_location(&self) -> SourceLocation {
        self.source_location
    }
    pub fn port_kind(&self) -> PortKind {
        T::port_kind(self.port_kind)
    }
    pub fn mem_name(&self) -> ScopedNameId {
        self.mem_name
    }
    pub fn port_index(&self) -> usize {
        self.port_index
    }
    pub fn port_name(&self) -> PortName {
        PortName {
            kind: self.port_kind(),
            index: self.port_index,
        }
    }
    pub fn mem_element_type(&self) -> Interned<dyn DynCanonicalType> {
        self.mem_element_type
    }
    pub fn addr_type(&self) -> DynUIntType {
        self.addr_type
    }
    pub fn canonical(&self) -> MemPort<DynPortType> {
        let Self {
            mem_name,
            source_location,
            port_kind,
            port_index,
            addr_type,
            mem_element_type,
        } = *self;
        MemPort::<DynPortType> {
            mem_name,
            source_location,
            port_kind: T::port_kind(port_kind),
            port_index,
            addr_type,
            mem_element_type,
        }
    }
    pub fn from_canonical(port: MemPort<DynPortType>) -> Self
    where
        T: PortStruct,
        T::Type: BundleType<Value = T>,
    {
        let MemPort {
            mem_name,
            source_location,
            port_kind,
            port_index,
            addr_type,
            mem_element_type,
        } = port;
        assert_eq!(port_kind, T::PORT_KIND, "port kind mismatch");
        Self {
            mem_name,
            source_location,
            port_kind: (),
            port_index,
            addr_type,
            mem_element_type,
        }
    }
    #[track_caller]
    pub fn new_unchecked(
        mem_name: ScopedNameId,
        source_location: SourceLocation,
        port_name: PortName,
        addr_type: DynUIntType,
        mem_element_type: Interned<dyn DynCanonicalType>,
    ) -> Self {
        assert!(
            mem_element_type.is_storable(),
            "memory element type must be a storable type"
        );
        Self {
            mem_name,
            source_location,
            port_kind: T::from_port_kind(port_name.kind),
            port_index: port_name.index,
            addr_type,
            mem_element_type,
        }
    }
    pub fn flow(&self) -> Flow {
        Flow::Sink
    }
    pub fn must_connect_to(&self) -> bool {
        true
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub enum ReadUnderWrite {
    Old,
    New,
    Undefined,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
struct MemImpl<T: ArrayTypeTrait, P> {
    scoped_name: ScopedNameId,
    source_location: SourceLocation,
    array_type: T,
    initial_value: Option<Interned<BitSlice>>,
    ports: P,
    read_latency: usize,
    write_latency: NonZeroU32,
    read_under_write: ReadUnderWrite,
    port_annotations: Interned<[TargetedAnnotation]>,
    mem_annotations: Interned<[Annotation]>,
}

pub struct Mem<VA: ValueArrayOrSlice + ?Sized>(
    Interned<MemImpl<ArrayType<VA>, Interned<[Interned<MemPort<DynPortType>>]>>>,
);

struct PortsDebug<'a>(&'a [Interned<MemPort<DynPortType>>]);

impl fmt::Debug for PortsDebug<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_map()
            .entries(
                self.0
                    .iter()
                    .map(|p| (p.port_name(), DebugAsDisplay("..."))),
            )
            .finish()
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> fmt::Debug for Mem<VA> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let MemImpl {
            scoped_name,
            source_location: _,
            array_type,
            initial_value,
            ports,
            read_latency,
            write_latency,
            read_under_write,
            port_annotations,
            mem_annotations,
        } = &*self.0;
        f.debug_struct("Mem")
            .field("name", scoped_name)
            .field("array_type", array_type)
            .field("initial_value", initial_value)
            .field("read_latency", read_latency)
            .field("write_latency", write_latency)
            .field("read_under_write", read_under_write)
            .field("ports", &PortsDebug(ports))
            .field("port_annotations", port_annotations)
            .field("mem_annotations", mem_annotations)
            .finish_non_exhaustive()
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Hash for Mem<VA> {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.0.hash(state);
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Eq for Mem<VA> {}

impl<VA: ValueArrayOrSlice + ?Sized> PartialEq for Mem<VA> {
    fn eq(&self, other: &Self) -> bool {
        self.0 == other.0
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Copy for Mem<VA> {}

impl<VA: ValueArrayOrSlice + ?Sized> Clone for Mem<VA> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Mem<VA> {
    #[allow(clippy::too_many_arguments)]
    #[track_caller]
    pub fn new_unchecked(
        scoped_name: ScopedNameId,
        source_location: SourceLocation,
        array_type: ArrayType<VA>,
        initial_value: Option<Interned<BitSlice>>,
        ports: Interned<[Interned<MemPort<DynPortType>>]>,
        read_latency: usize,
        write_latency: NonZeroU32,
        read_under_write: ReadUnderWrite,
        port_annotations: Interned<[TargetedAnnotation]>,
        mem_annotations: Interned<[Annotation]>,
    ) -> Self {
        if let Some(initial_value) = initial_value {
            MemBuilder::<VA>::check_initial_value_bit_slice(
                array_type.element(),
                Some(array_type.len()),
                initial_value,
            );
        }
        let addr_width = memory_addr_width(array_type.len());
        let expected_mem_element_type = array_type.element().canonical_dyn();
        assert!(
            expected_mem_element_type.is_storable(),
            "memory element type must be a storable type"
        );
        for (index, port) in ports.iter().enumerate() {
            let MemPort {
                mem_name,
                source_location: _,
                port_kind: _,
                port_index,
                addr_type,
                mem_element_type,
            } = **port;
            assert_eq!(mem_name, scoped_name, "memory name must match with ports");
            assert_eq!(
                port_index, index,
                "port index must match position in ports list"
            );
            assert_eq!(
                addr_type.width, addr_width,
                "memory address width must match with ports"
            );
            assert_eq!(
                mem_element_type, expected_mem_element_type,
                "memory element type must match with ports"
            );
        }
        for port_annotation in port_annotations {
            let base = port_annotation.target().base();
            let Some(port) = base.mem_port() else {
                panic!("port annotation must have a memory port as its target base");
            };
            assert_eq!(
                Some(port),
                ports.get(port.port_index).map(|v| &**v),
                "port on memory must match annotation's target base"
            );
        }
        Self(Intern::intern_sized(MemImpl {
            scoped_name,
            source_location,
            array_type,
            initial_value,
            ports,
            read_latency,
            write_latency,
            read_under_write,
            port_annotations,
            mem_annotations,
        }))
    }
    pub fn scoped_name(self) -> ScopedNameId {
        self.0.scoped_name
    }
    pub fn source_location(self) -> SourceLocation {
        self.0.source_location
    }
    pub fn array_type(self) -> ArrayType<VA> {
        self.0.array_type.clone()
    }
    pub fn initial_value(self) -> Option<Interned<BitSlice>> {
        self.0.initial_value
    }
    pub fn ports(self) -> Interned<[Interned<MemPort<DynPortType>>]> {
        self.0.ports
    }
    pub fn read_latency(self) -> usize {
        self.0.read_latency
    }
    pub fn write_latency(self) -> NonZeroU32 {
        self.0.write_latency
    }
    pub fn read_under_write(self) -> ReadUnderWrite {
        self.0.read_under_write
    }
    pub fn port_annotations(self) -> Interned<[TargetedAnnotation]> {
        self.0.port_annotations
    }
    pub fn mem_annotations(self) -> Interned<[Annotation]> {
        self.0.mem_annotations
    }
    pub fn canonical(self) -> Mem<[DynCanonicalValue]> {
        let MemImpl {
            scoped_name,
            source_location,
            ref array_type,
            initial_value,
            ports,
            read_latency,
            write_latency,
            read_under_write,
            port_annotations,
            mem_annotations,
        } = *self.0;
        let array_type = array_type.canonical();
        Mem(Intern::intern_sized(MemImpl {
            scoped_name,
            source_location,
            array_type,
            initial_value,
            ports,
            read_latency,
            write_latency,
            read_under_write,
            port_annotations,
            mem_annotations,
        }))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Default)]
struct MaybeSpecified<T>(Option<T>);

impl<T: fmt::Debug> fmt::Debug for MaybeSpecified<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(v) = &self.0 {
            v.fmt(f)
        } else {
            f.write_str("<not yet specified>")
        }
    }
}

pub(crate) struct MemBuilderTarget {
    pub(crate) scoped_name: ScopedNameId,
    pub(crate) source_location: SourceLocation,
    pub(crate) mem_element_type: Interned<dyn DynCanonicalType>,
    pub(crate) depth: Option<usize>,
    pub(crate) initial_value: Option<Interned<BitSlice>>,
    pub(crate) ports: Vec<Interned<MemPort<DynPortType>>>,
    pub(crate) read_latency: usize,
    pub(crate) write_latency: NonZeroU32,
    pub(crate) read_under_write: ReadUnderWrite,
    pub(crate) port_annotations: Vec<TargetedAnnotation>,
    pub(crate) mem_annotations: Vec<Annotation>,
}

impl MemBuilderTarget {
    pub(crate) fn make_memory(&self) -> Option<Mem<[DynCanonicalValue]>> {
        Some(Mem::new_unchecked(
            self.scoped_name,
            self.source_location,
            ArrayType::new_slice(self.mem_element_type, self.depth?),
            self.initial_value,
            Intern::intern(&self.ports),
            self.read_latency,
            self.write_latency,
            self.read_under_write,
            Intern::intern(&self.port_annotations),
            Intern::intern(&self.mem_annotations),
        ))
    }
    fn debug_fmt(
        &self,
        struct_name: &str,
        mem_element_type: &dyn fmt::Debug,
        f: &mut fmt::Formatter<'_>,
    ) -> fmt::Result {
        let Self {
            scoped_name,
            source_location: _,
            mem_element_type: _,
            depth,
            initial_value,
            ports,
            read_latency,
            write_latency,
            read_under_write,
            port_annotations,
            mem_annotations,
        } = self;
        f.debug_struct(struct_name)
            .field("name", scoped_name)
            .field("mem_element_type", mem_element_type)
            .field("depth", &MaybeSpecified(*depth))
            .field("initial_value", initial_value)
            .field("read_latency", read_latency)
            .field("write_latency", write_latency)
            .field("read_under_write", read_under_write)
            .field("ports", ports)
            .field("port_annotations", port_annotations)
            .field("mem_annotations", mem_annotations)
            .finish_non_exhaustive()
    }
}

impl fmt::Debug for MemBuilderTarget {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.debug_fmt("MemBuilderTarget", &self.mem_element_type, f)
    }
}

pub struct MemBuilder<VA: ValueArrayOrSlice + ?Sized> {
    mem_element_type: VA::ElementType,
    target: Rc<RefCell<MemBuilderTarget>>,
}

impl<VA: ValueArrayOrSlice + ?Sized> fmt::Debug for MemBuilder<VA> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            mem_element_type,
            target,
        } = &self;
        target.borrow().debug_fmt("MemBuilder", mem_element_type, f)
    }
}

pub fn memory_addr_width(depth: usize) -> usize {
    depth
        .checked_next_power_of_two()
        .map_or(usize::BITS, usize::ilog2) as usize
}

impl<VA: ValueArrayOrSlice + ?Sized> MemBuilder<VA> {
    #[track_caller]
    fn check_initial_value_bit_slice(
        mem_element_type: &VA::ElementType,
        depth: Option<usize>,
        initial_value: Interned<BitSlice>,
    ) -> Interned<BitSlice> {
        if let Some(depth) = depth {
            let expected_len = depth.checked_mul(mem_element_type.bit_width()).expect(
                "memory must be small enough that its initializer bit length fits in usize",
            );
            assert_eq!(
                expected_len,
                initial_value.len(),
                "Mem's initializer bit length doesn't match the expected value",
            );
        }
        assert!(
            initial_value
                .len()
                .checked_rem(mem_element_type.bit_width())
                .unwrap_or(initial_value.len())
                == 0,
            "Mem's initializer bit length must be a multiple of the element type's bit width",
        );
        initial_value
    }
    #[track_caller]
    fn check_initial_value_expr(
        mem_element_type: &VA::ElementType,
        depth: Option<usize>,
        initial_value: Expr<Array<[DynCanonicalValue]>>,
    ) -> Interned<BitSlice> {
        let initial_value_ty = initial_value.canonical_type();
        assert_eq!(
            *mem_element_type,
            <VA::ElementType as DynType>::from_dyn_canonical_type(*initial_value_ty.element()),
            "Mem's element type must match initializer's element type",
        );
        if let Some(depth) = depth {
            assert_eq!(
                depth,
                initial_value_ty.len(),
                "Mem's depth must match initializer's length"
            );
        }
        let Ok(retval) = initial_value.to_literal_bits() else {
            panic!("Mem's initializer must be convertible to literal bits");
        };
        debug_assert_eq!(
            retval.len(),
            initial_value_ty.bit_width(),
            "initial value produced wrong literal bits length"
        );
        retval
    }
    #[track_caller]
    pub(crate) fn new(
        scoped_name: ScopedNameId,
        source_location: SourceLocation,
        mem_element_type: VA::ElementType,
    ) -> (Self, Rc<RefCell<MemBuilderTarget>>) {
        let canonical_mem_element_type = mem_element_type.canonical_dyn();
        assert!(
            canonical_mem_element_type.is_storable(),
            "memory element type must be a storable type"
        );
        let target = Rc::new(RefCell::new(MemBuilderTarget {
            scoped_name,
            source_location,
            mem_element_type: canonical_mem_element_type,
            depth: VA::FIXED_LEN_TYPE.map(VA::len_from_len_type),
            initial_value: None,
            ports: vec![],
            read_latency: 0,
            write_latency: NonZeroU32::new(1).unwrap(),
            read_under_write: ReadUnderWrite::Old,
            port_annotations: vec![],
            mem_annotations: vec![],
        }));
        (
            Self {
                mem_element_type,
                target: Rc::clone(&target),
            },
            target,
        )
    }
    #[track_caller]
    fn new_port_impl(
        &mut self,
        source_location: SourceLocation,
        port_kind: PortKind,
    ) -> Interned<MemPort<DynPortType>> {
        let mut target = self.target.borrow_mut();
        let Some(depth) = target.depth else {
            panic!("MemBuilder::depth must be called before adding ports");
        };
        let port = Intern::intern_sized(MemPort {
            mem_name: target.scoped_name,
            source_location,
            port_kind,
            port_index: target.ports.len(),
            addr_type: DynUIntType::new(memory_addr_width(depth)),
            mem_element_type: target.mem_element_type,
        });
        target.ports.push(port);
        port
    }
    #[track_caller]
    pub fn new_port_with_loc(
        &mut self,
        source_location: SourceLocation,
        kind: PortKind,
    ) -> Expr<DynBundle> {
        Expr::new_unchecked(ExprEnum::MemPort(self.new_port_impl(source_location, kind)))
    }
    #[track_caller]
    pub fn new_port(&mut self, kind: PortKind) -> Expr<DynBundle> {
        self.new_port_with_loc(SourceLocation::caller(), kind)
    }
    #[track_caller]
    pub fn new_read_port_with_loc(
        &mut self,
        source_location: SourceLocation,
    ) -> Expr<ReadStruct<VA::Element>> {
        Expr::new_unchecked(ExprEnum::MemPort(
            self.new_port_impl(source_location, PortKind::ReadOnly),
        ))
    }
    #[track_caller]
    pub fn new_read_port(&mut self) -> Expr<ReadStruct<VA::Element>> {
        self.new_read_port_with_loc(SourceLocation::caller())
    }
    #[track_caller]
    pub fn new_write_port_with_loc(
        &mut self,
        source_location: SourceLocation,
    ) -> Expr<WriteStruct<VA::Element>> {
        Expr::new_unchecked(ExprEnum::MemPort(
            self.new_port_impl(source_location, PortKind::WriteOnly),
        ))
    }
    #[track_caller]
    pub fn new_write_port(&mut self) -> Expr<WriteStruct<VA::Element>> {
        self.new_write_port_with_loc(SourceLocation::caller())
    }
    #[track_caller]
    pub fn new_rw_port_with_loc(
        &mut self,
        source_location: SourceLocation,
    ) -> Expr<ReadWriteStruct<VA::Element>> {
        Expr::new_unchecked(ExprEnum::MemPort(
            self.new_port_impl(source_location, PortKind::ReadWrite),
        ))
    }
    #[track_caller]
    pub fn new_rw_port(&mut self) -> Expr<ReadWriteStruct<VA::Element>> {
        self.new_rw_port_with_loc(SourceLocation::caller())
    }
    pub fn scoped_name(&self) -> ScopedNameId {
        self.target.borrow().scoped_name
    }
    pub fn source_location(&self) -> SourceLocation {
        self.target.borrow().source_location
    }
    pub fn get_mem_element_type(&self) -> &VA::ElementType {
        &self.mem_element_type
    }
    #[allow(clippy::result_unit_err)]
    pub fn get_depth(&self) -> Result<usize, ()> {
        self.target.borrow().depth.ok_or(())
    }
    #[track_caller]
    pub fn depth(&mut self, depth: usize) {
        let mut target = self.target.borrow_mut();
        assert!(
            target.ports.is_empty(),
            "Mem depth can only be set before adding ports"
        );
        if let Some(expected) = target.depth {
            assert_eq!(
                expected, depth,
                "Mem depth doesn't match previously specified value"
            );
        }
        target.depth = Some(depth);
    }
    #[allow(clippy::result_unit_err)]
    pub fn get_array_type(&self) -> Result<ArrayType<VA>, ()> {
        Ok(ArrayType::new_with_len(
            self.mem_element_type.clone(),
            self.get_depth()?,
        ))
    }
    pub fn get_initial_value(&self) -> Option<Interned<BitSlice>> {
        self.target.borrow().initial_value
    }
    #[track_caller]
    pub fn initial_value(&mut self, initial_value: impl ToExpr<Type = ArrayType<VA>>) {
        let mut target = self.target.borrow_mut();
        if target.initial_value.is_some() {
            panic!("can't set Mem's initial value more than once");
        }
        let initial_value = initial_value.to_expr().canonical();
        target.initial_value = Some(Self::check_initial_value_expr(
            &self.mem_element_type,
            target.depth,
            initial_value,
        ));
        target.depth = Some(initial_value.ty().len());
    }
    #[track_caller]
    pub fn initial_value_bit_slice(&mut self, initial_value: Interned<BitSlice>) {
        let mut target = self.target.borrow_mut();
        if target.initial_value.is_some() {
            panic!("can't set Mem's initial value more than once");
        }
        target.initial_value = Some(Self::check_initial_value_bit_slice(
            &self.mem_element_type,
            target.depth,
            initial_value,
        ));
        let element_bit_width = self.mem_element_type.bit_width();
        if element_bit_width != 0 {
            target.depth = Some(initial_value.len() / element_bit_width);
        }
    }
    pub fn get_read_latency(&self) -> usize {
        self.target.borrow().read_latency
    }
    pub fn read_latency(&mut self, read_latency: usize) {
        self.target.borrow_mut().read_latency = read_latency;
    }
    pub fn get_write_latency(&self) -> NonZeroU32 {
        self.target.borrow().write_latency
    }
    pub fn write_latency(&mut self, write_latency: NonZeroU32) {
        self.target.borrow_mut().write_latency = write_latency;
    }
    pub fn get_read_under_write(&self) -> ReadUnderWrite {
        self.target.borrow().read_under_write
    }
    pub fn read_under_write(&mut self, read_under_write: ReadUnderWrite) {
        self.target.borrow_mut().read_under_write = read_under_write;
    }
    #[track_caller]
    pub fn annotate(&mut self, annotations: impl IntoAnnotations) {
        self.target
            .borrow_mut()
            .mem_annotations
            .extend(annotations.into_annotations());
    }
}
