// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    expr::{Expr, ExprTrait, Flow, ToExpr},
    intern::Interned,
    module::{NameId, ScopedNameId},
    source_location::SourceLocation,
    ty::{DynCanonicalType, DynType, Type},
};
use std::fmt;

#[derive(Clone, Eq, PartialEq, Hash)]
pub struct Wire<T: Type> {
    name: ScopedNameId,
    source_location: SourceLocation,
    ty: T,
}

impl<T: Type + fmt::Debug> fmt::Debug for Wire<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("Wire")
            .field("name", &self.name)
            .field("ty", &self.ty)
            .finish_non_exhaustive()
    }
}

impl<T: Type> ToExpr for Wire<T> {
    type Type = T;

    fn ty(&self) -> Self::Type {
        self.ty.clone()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: Type> Wire<T> {
    pub fn canonical(&self) -> Wire<T::CanonicalType> {
        let Self {
            name,
            source_location,
            ref ty,
        } = *self;
        Wire {
            name,
            source_location,
            ty: ty.canonical(),
        }
    }
    pub fn to_dyn_wire(&self) -> Wire<Interned<dyn DynType>> {
        let Self {
            name,
            source_location,
            ref ty,
        } = *self;
        Wire {
            name,
            source_location,
            ty: ty.to_dyn(),
        }
    }
    pub fn to_dyn_canonical_wire(&self) -> Wire<Interned<dyn DynCanonicalType>> {
        let Self {
            name,
            source_location,
            ref ty,
        } = *self;
        Wire {
            name,
            source_location,
            ty: ty.canonical_dyn(),
        }
    }
    pub fn new_unchecked(
        scoped_name: ScopedNameId,
        source_location: SourceLocation,
        ty: T,
    ) -> Self {
        Self {
            name: scoped_name,
            source_location,
            ty,
        }
    }
    pub fn source_location(&self) -> SourceLocation {
        self.source_location
    }
    pub fn containing_module_name(&self) -> Interned<str> {
        self.containing_module_name_id().0
    }
    pub fn containing_module_name_id(&self) -> NameId {
        self.name.0
    }
    pub fn name(&self) -> Interned<str> {
        self.name_id().0
    }
    pub fn name_id(&self) -> NameId {
        self.name.1
    }
    pub fn scoped_name(&self) -> ScopedNameId {
        self.name
    }
    pub fn flow(&self) -> Flow {
        Flow::Duplex
    }
    pub fn must_connect_to(&self) -> bool {
        true
    }
}
